Function ATan2(y As Double, x As Double) As Double
'
' Returns arctangent (y/x) with results from 0 to 360; also tolerates x = 0
'
  Dim Pi As Double
  Pi = 3.14159265359
  
  If x = 0 Then
    If y > 0 Then
      ATan2 = Pi / 2
    Else
      ATan2 = -Pi / 2
    End If
  Else
    ATan2 = Atn(y / x)
    If x < 0 Then
      If y >= 0 Then
        ATan2 = ATan2 + Pi
      Else
        ATan2 = ATan2 - Pi
      End If
    End If
  End If
  If ATan2 < 0 Then ATan2 = ATan2 + 2 * Pi
End Function

Function MH2Lat(mh_in)
'/   returns the latitude of the Maidenhead location
'/   the Maidenhead input string must be in lower case
    Dim mhsz As Integer
    Dim lt As Double
    Dim mh As String
    
    mh = LCase(mh_in)
  
    mhsz = Len(mh)
    If (mhsz <> 6 And mhsz <> 4) Then GoTo mhl_error
    
    If (Mid(mh, 1, 1) < "a" Or Mid(mh, 1, 1) > "r") Then GoTo mhl_error
    If (Mid(mh, 2, 1) < "a" Or Mid(mh, 2, 1) > "s") Then GoTo mhl_error
    If (Mid(mh, 3, 1) < "0" Or Mid(mh, 3, 1) > "9") Then GoTo mhl_error
    If (Mid(mh, 4, 1) < "0" Or Mid(mh, 4, 1) > "9") Then GoTo mhl_error
    
    If mhsz = 6 Then
        If Mid(mh, 5, 1) < "a" Or Mid(mh, 5, 1) > "x" Then GoTo mhl_error
        If Mid(mh, 6, 1) < "a" Or Mid(mh, 6, 1) > "x" Then GoTo mhl_error
    End If
    lt = 10# * CDbl(Asc(Mid(mh, 2, 1)) - Asc("a"))
    lt = lt + CDbl(1# * (Asc(Mid(mh, 4, 1)) - Asc("0")))
    If (mhsz = 6) Then
        lt = lt + (0.04166666667 * CDbl(Asc(Mid(mh, 6, 1)) - Asc("a")))
    End If
    
'/  adjust for center of grid
    If (mhsz = 6) Then lt = lt + 0.02083333333 _
    Else lt = lt + 0.5

    lt = lt - 90
    
    MH2Lat = lt
    Exit Function
mhl_error:
    MH2Lat = ""
End Function

Function MH2Lon(mh_in)
'/   returns the latitude of the Maidenhead location
'/   the Maidenhead input string must be in lower case
    Dim mhsz As Integer
    Dim ln As Double
    Dim mh As String
  
    mh = LCase(mh_in)
  
    mhsz = Len(mh)
    If (mhsz <> 6 And mhsz <> 4) Then GoTo mhl_error
    
    If (Mid(mh, 1, 1) < "a" Or Mid(mh, 1, 1) > "r") Then GoTo mhl_error
    If (Mid(mh, 2, 1) < "a" Or Mid(mh, 2, 1) > "s") Then GoTo mhl_error
    If (Mid(mh, 3, 1) < "0" Or Mid(mh, 3, 1) > "9") Then GoTo mhl_error
    If (Mid(mh, 4, 1) < "0" Or Mid(mh, 4, 1) > "9") Then GoTo mhl_error
    
    If mhsz = 6 Then
        If Mid(mh, 5, 1) < "a" Or Mid(mh, 5, 1) > "x" Then GoTo mhl_error
        If Mid(mh, 6, 1) < "a" Or Mid(mh, 6, 1) > "x" Then GoTo mhl_error
    End If
    
    ln = 20# * CDbl(Asc(Mid(mh, 1, 1)) - Asc("a"))
    ln = ln + 2# * CDbl(Asc(Mid(mh, 3, 1)) - Asc("0"))
    If (mhsz = 6) Then
        ln = ln + (0.08333333333 * CDbl(Asc(Mid(mh, 5, 1)) - Asc("a")))
    End If
    
'/  adjust for center of grid
    If (mhsz = 6) Then ln = ln + 0.04166666667 _
    Else ln = ln + 1#

    ln = ln - 180
    
    MH2Lon = ln
    Exit Function
mhl_error:
    MH2Lon = ""
End Function
Function LatLon2MH(Lat As Double, Lon As Double) As String
    Dim mh As String
    Dim lt1 As Double
    Dim lg1 As Double
    Dim wrk As Double, wrk1 As Double
    Dim i As Integer

    mh = " "

    '/ adjust lt/lng for mh with special for wrapping boundry cases
    If (Lon >= 180) Then
        lg1 = 359.999999999
    Else: lg1 = Lon + 180#  '/ convert degree longitude for maidenhead */
    End If
    If (Lat >= 90) Then
        lt1 = 179.999999999
    Else: lt1 = Lat + 90#   '/ convert degree latitude for maidenhead  */
    End If

    mh = Chr(Asc("a") + Int(lg1 / 20))
    mh = mh & Chr(Asc("a") + Int(lt1 / 10))
    
    wrk = lg1 - (20# * CDbl(Int(lg1 / 20)))
    mh = mh & Chr(Asc("0") + Int(wrk / 2))
    wrk1 = lt1 - (10# * CDbl(Int(lt1 / 10)))
    mh = mh & Chr(Asc("0") + Int(wrk1))
    
    wrk = wrk - (2 * Int(wrk / 2))
    mh = mh & Chr(Asc("a") + Int(wrk * 60# / 5#))
    wrk1 = wrk1 - Int(wrk1)
    mh = mh & Chr(Asc("a") + Int(wrk1 * 60# / 2.5))
    
    LatLon2MH = mh
End Function
Function Distance(lat1 As Double, lon1 As Double, lat2 As Double, lon2 As Double) As Double

    Dim th1 As Double, th2 As Double, thm As Double, dthm As Double
    Dim dlamm As Double, dlam As Double, sindlamm As Double
    Dim costhm As Double, sinthm As Double, cosdthm As Double
    Dim sindthm As Double, L As Double, E As Double, cosd As Double
    Dim D As Double, x As Double, y As Double, T As Double
    Dim sind As Double, DD As Double
    Dim AA As Double, B As Double
    Dim Val As Double
    Const DTOL As Double = 1E-16
    '/ Values for WGS84 world model
    Const a As Double = 6378137#
    Const onef As Double = 0.996647189335
    Const f4 As Double = 0.0008382026662
    Const f64 As Double = 1.756459274E-07

    If (Not (IsNumeric(lat1) And IsNumeric(lon1) And IsNumeric(lat2) And IsNumeric(lon2))) Then
        Distance = ""
        Exit Function
    End If
    
    '/ from degrees to radians
    lat1 = lat1 * 1.74532925199433E-02
    lon1 = lon1 * 1.74532925199433E-02
    lat2 = lat2 * 1.74532925199433E-02
    lon2 = lon2 * 1.74532925199433E-02

    '/ get geocentric latitude th means theta
    th1 = Atn(onef * Tan(lat1))
    th2 = Atn(onef * Tan(lat2))

    thm = 0.5 * (th1 + th2)
    dthm = 0.5 * (th2 - th1)
    
    dlam = lon2 - lon1
    '/ adjlon inline
    While (Abs(dlam) > 3.14159265359)
        If dlam < 0# Then dlam = dlam + 6.28318530717959 Else _
        dlam = dlam - 6.28318530717959
    Wend
        
    dlamm = 0.5 * dlam

    sindlamm = Sin(dlamm)
    costhm = Cos(thm)
    sinthm = Sin(thm)
    cosdthm = Cos(dthm)
    sindthm = Sin(dthm)
    
    L = sindthm * sindthm + (cosdthm * cosdthm - sinthm * sinthm) _
        * sindlamm * sindlamm

    If ((1 - L) < DTOL) Then L = 1# - DTOL

    cosd = 1# - L - L

    '/ test for coincidence
    '/ minimum distance ~ 1 km
    If (cosd > 0.9999999883) Then
        '/ loc1 and loc2 coincide
        Distance = 0#
        Exit Function
    End If
    

    '/ d is the Azimuth angle of the great circle arc
    '/ compute equivalent of... D = Acos(cosd)
    D = Atn(-cosd / Sqr(-cosd * cosd + 1)) + 2# * Atn(1)

    '/ compute distance (S) and
    E = cosd + cosd
    sind = Sin(D)
    y = sinthm * cosdthm
    y = y * ((y + y) / (1# - L))
    T = sindthm * costhm
    T = T * ((T + T) / L)
    x = y + T
    y = y - T
    T = D / sind
    DD = 4# * T * T
    AA = DD * E
    B = DD + DD

        
    Val = y * (B + E * y) + DD * x * y
    Val = x * (AA + (T - 0.5 * (AA - E)) * x) - Val
    Val = f64 * Val
    Val = T - f4 * (T * x - y) + Val
    Val = a * sind * Val
       
    Val = Val / 1000#
        
    Distance = Val

End Function

Function Bearing(lat1 As Double, lon1 As Double, lat2 As Double, lon2 As Double) As Double
    Dim th1 As Double, th2 As Double, thm As Double, dthm As Double
    Dim dlamm As Double, dlam As Double, sindlamm As Double
    Dim costhm As Double, sinthm As Double, cosdthm As Double
    Dim sindthm As Double, L As Double, E As Double, cosd As Double
    Dim D As Double, x As Double, y As Double, T As Double
    Dim sind As Double, DD As Double
    Dim AA As Double, B As Double
    Dim u As Double, v As Double
    Dim Val As Double
    Dim antip As Integer
    Const DTOL As Double = 1E-16
    '/ Values for WGS84 world model
    Const a As Double = 6378137#
    Const onef As Double = 0.996647189335
    Const f2 As Double = 0.0016764053324
    Const f4 As Double = 0.0008382026662
    Const f64 As Double = 1.756459274E-07
    
    If (Not (IsNumeric(lat1) And IsNumeric(lon1) And IsNumeric(lat2) And IsNumeric(lon2))) Then
        Bearing = ""
        Exit Function
    End If
    
    '/ from degrees to radians
    lat1 = lat1 * 1.74532925199433E-02
    lon1 = lon1 * 1.74532925199433E-02
    lat2 = lat2 * 1.74532925199433E-02
    lon2 = lon2 * 1.74532925199433E-02

    '/ get geocentric latitude th means theta
    th1 = Atn(onef * Tan(lat1))
    th2 = Atn(onef * Tan(lat2))

    thm = 0.5 * (th1 + th2)
    dthm = 0.5 * (th2 - th1)
    
    dlam = lon2 - lon1
    '/ adjlon inline
    While (Abs(dlam) > 3.14159265359)
        If dlam < 0# Then dlam = dlam + 6.28318530717959 Else _
        dlam = dlam - 6.28318530717959
    Wend
        
    dlamm = 0.5 * dlam

    sindlamm = Sin(dlamm)
    costhm = Cos(thm)
    sinthm = Sin(thm)
    cosdthm = Cos(dthm)
    sindthm = Sin(dthm)
    
    L = sindthm * sindthm + (cosdthm * cosdthm - sinthm * sinthm) _
        * sindlamm * sindlamm

    If ((1 - L) < DTOL) Then L = 1 - DTOL

    cosd = 1 - L - L

    '/ test for coincidence
    '/ minimum distance ~ 1 km
    If (cosd > 0.9999999883) Then
        '/ loc1 and loc2 coincide
        Bearing = 0#
        Exit Function
    End If
    

    '/ d is the Azimuth angle of the great circle arc
    '/ compute equivalent of... D = Acos(cosd)
    D = Atn(-cosd / Sqr(-cosd * cosd + 1)) + 2 * Atn(1)

    '/ compute distance (S) and
    E = cosd + cosd
    sind = Sin(D)
    y = sinthm * cosdthm
    y = y * ((y + y) / (1# - L))
    T = sindthm * costhm
    T = T * ((T + T) / L)
    x = y + T
    y = y - T
    T = D / sind
    DD = 4# * T * T
    AA = DD * E
    B = DD + DD
    
    tandlammp = Tan(0.5 * (dlam - 0.25 * (y + y - E * (4# - x)) * _
        (f2 * T + f64 * (32# * T - (20# * T - AA) _
        * x - (B + 4#) * y)) * Tan(dlam)))
            
    If (antip) Then
        '/ for antipod return azimuths > 2pi as undefined flag
        Bearing = 555
        Exit Function
    End If
    
    u = ATan2(sindthm, (tandlammp * costhm))
    v = ATan2(cosdthm, (tandlammp * sinthm))
    
    '/ get direct azimuth
    Val = 6.28318530717959 + v - u
    While (Abs(Val) > 3.14159265359)
        If Val < 0# Then Val = Val + 6.28318530717959 Else _
        Val = Val - 6.28318530717959
    Wend
    
    Val = Val * 57.2957795130823
    If Val < 0# Then Val = Val + 360#
    Bearing = Val

End Function
